open Bigarray
open Printf

<<
#include <cblas.h>
>>

module Vector =
struct
  type t = (float, float64_elt, c_layout) Array1.t
  let zeros n = Array1.create float64 c_layout n
  let of_array v = Array1.of_array float64 c_layout v
  let of_list v = Array1.of_array float64 c_layout (Array.of_list v)
end

module Matrix =
struct
  type t = (float, float64_elt, c_layout) Array2.t
  let zeros n m = Array2.create float64 c_layout n m
  let of_array m = Array2.of_array float64 c_layout m
  let print m =
    let n1 = Array2.dim1 m in
    let n2 = Array2.dim2 m in
      for i = 0 to n1 - 1 do
        print_string "( ";
        for j = 0 to n2 - 1 do
          printf "%f " m.{i, j}
        done;
      print_endline ")"
      done;
end
type vector = Vector.t
type matrix = Matrix.t

exception Size_mismatch
exception Incompatible_shapes
Callback.register_exception "Size_mismatch" Size_mismatch;
Callback.register_exception "Incompatible_shapes" Incompatible_shapes;
register_fromval "identity" "value": vector
register_fromval "identity" "value": matrix

letext dot (u:vector) (v:vector) : float = <<
    double r;
    vector(u, n_u, double* data_u);
    vector(v, n_v, double* data_v);

    if (n_u != n_v)
      raise("Size_mismatch");

    r = cblas_ddot(n_u, data_u, 1, data_v, 1);
    Return(r);
>>

letext gemm (alpha:float) (a:matrix) (b:matrix) (beta:float) (c:matrix): unit = <<
    matrix(a, n1_a, n2_a, double* data_a);
    matrix(b, n1_b, n2_b, double* data_b);
    matrix(c, n1_c, n2_c, double* data_c);

    if (n2_a != n1_b || n1_a != n1_c || n2_b != n2_c)
      raise("Incompatible_shapes");

    cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, n1_a, n2_a, n2_b, alpha, data_a, n1_a, data_b, n1_b, beta, data_c, n1_c);
    Return();
>>

let vector_dot () =
  let u = Vector.of_list [ 1.; 1.; 1.; 1. ] in
  let v = Vector.of_array [| 1.; 2.; 1.; 1. |] in
  let w = Vector.zeros 5 in
    Printf.printf "%f\n" (dot u v);
    try ignore (dot u w) with Size_mismatch -> Printf.printf "Size_mismatch !\n";
    ()

let matrix_prod () =
  let a = Matrix.of_array [| [| 1. |]; [| 1. |] |] in
  let b = Matrix.of_array [| [| 1.; 1. |] |] in
  let c = Matrix.zeros 2 2 in
    gemm 1. a b 0. c;
    Matrix.print c;
    ()

let _ =
  vector_dot ();
  matrix_prod ();
  ()

